/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "COPYING"). If not, If not, see <http://www.gnu.org/licenses/>.        */
/*											*/
/*--------------------------------------------------------------------------------------*/
/*											*/
/*  Copyright   Joerg Anders, TU Chemnitz, Fakultaet fuer Informatik, GERMANY           */
/*		ja@informatik.tu-chemnitz.de						*/
/*											*/
/*											*/
/****************************************************************************************/

#include <math.h>
#include "freetext.h"
#include "chordorrest.h"
#include "textdialog.h"
#include "staff.h"
#include "system.h"
#include "page.h"

#define X_POS_PAGE_REL(p) ((m_element->getPage()->getContetXpos() + (p)) * zoom_factor - leftx)
#define Y_POS_PAGE_REL(p) ((p) * zoom_factor - topy)
#define X_POS_PAGE_REL_INVERSE(p) (((p) + leftx) / zoom_factor - m_element->getPage()->getContetXpos()) 
#define Y_POS_PAGE_REL_INVERSE(p) (((p) + topy - (TOP_BOTTOM_BORDER + DEFAULT_BORDER)) / zoom_factor)

NedFreeText::NedFreeText(char *text, GtkWidget *drawing_area, int anchor,  unsigned int tempo, bool enable_tempo, unsigned int volume, bool enable_volume,
			unsigned short segno_sign, char *font_family, double font_size, cairo_font_slant_t font_slant, cairo_font_weight_t font_weight) : NedFreeReplaceable(),
		m_midi_volume(volume), m_enable_volume_change(enable_volume), m_enable_tempo_change(enable_tempo), m_segno_sign(segno_sign), 
			m_font_family(font_family), m_font_size(font_size), m_font_slant(font_slant), m_font_weight(font_weight), m_anchor(anchor), m_tempo(tempo),
			m_text(NULL), m_drawing_area(drawing_area) {
		if (text != NULL) {
			m_text = strdup(text);
		}
		computeBbox();
}

NedFreeText::NedFreeText(char *text, GtkWidget *drawing_area, int anchor,  unsigned int tempo, bool enable_tempo, unsigned int volume, bool enable_volume,
		unsigned short segno_sign, char *font_family, double font_size, cairo_font_slant_t font_slant, cairo_font_weight_t font_weight,
		NedChordOrRest *element) : NedFreeReplaceable(element), 
			m_midi_volume(volume), m_enable_volume_change(enable_volume), m_enable_tempo_change(enable_tempo),
			m_segno_sign(segno_sign), m_font_family(font_family),
			m_font_size(font_size), m_font_slant(font_slant), m_font_weight(font_weight), m_anchor(anchor), m_tempo(tempo),
			m_text(NULL), m_drawing_area(drawing_area) {
			m_text = strdup(text);
		if (text != NULL) {
			m_text = strdup(text);
		}
		computeBbox();
}

NedFreeText::NedFreeText(char *text, GtkWidget *drawing_area, int anchor,  unsigned int tempo, bool enable_tempo, unsigned int volume, bool enable_volume,
		unsigned short segno_sign, char *font_family, double font_size, cairo_font_slant_t font_slant, cairo_font_weight_t font_weight,
		NedChordOrRest *element, double x, double y, bool relative) :
			NedFreeReplaceable(element, x, y, relative), 
		m_midi_volume(volume), m_enable_volume_change(enable_volume), m_enable_tempo_change(enable_tempo), m_segno_sign(segno_sign),
			m_font_family(font_family), m_font_size(font_size),
			m_font_slant(font_slant), m_font_weight(font_weight), m_anchor(anchor), m_tempo(tempo),
			m_text(NULL), m_drawing_area(drawing_area) {
		if (text != NULL) {
			m_text = strdup(text);
		}
		computeBbox();
}


NedFreeReplaceable *NedFreeText::clone(struct addr_ref_str **slurlist, NedChordOrRest *newelement, GtkWidget *drawing_area /* = NULL */) {
	char *nfamily;

	if (drawing_area == NULL) {
		NedResource::Abort("NedFreeText::clone(1)");
	}
	if ((nfamily = (char *) g_try_malloc(strlen(m_font_family) + 1)) == NULL) {
		NedResource::Abort("NedFreeText::clone(2)");
	}
	strcpy(nfamily, m_font_family);
	NedFreeText *freetext = new NedFreeText(m_text, drawing_area, m_anchor, m_tempo, m_enable_tempo_change, m_midi_volume, m_enable_volume_change,
		m_segno_sign, nfamily, m_font_size, m_font_slant, m_font_weight, newelement, m_x, m_y, true);
	return freetext;
}

void NedFreeText::adjust_pointers(struct addr_ref_str *slurlist) {
}

void NedFreeText::startContextDialog(GtkWindow *ref) {
	bool state;
	char *newtext;
	gchar *newfont_family;
	cairo_font_slant_t newfont_slant;
	cairo_font_weight_t newfont_weight;
	bool newenable_volume;
	bool newenable_tempo;
	unsigned short newsegno_sign;
	int new_anchor;
	unsigned int newvolume;
	unsigned int newtempo;
	double newfont_size;

	NedTextDialog *dialog = new NedTextDialog (GTK_WINDOW(ref), m_text, m_anchor, m_tempo, m_enable_tempo_change, m_midi_volume, m_enable_volume_change,
		m_segno_sign, m_font_family, m_font_size, m_font_slant, m_font_weight);
	dialog->getValues(&state, &newtext, &new_anchor, &newtempo, &newenable_tempo, &newvolume, &newenable_volume, &newsegno_sign, &newfont_family, &newfont_size, &newfont_slant, &newfont_weight);
	if (state) {
		if (newtext != NULL) {
			if (strlen(newtext) > 0) {
			/*
				if (m_text != NULL) {
					free(m_text);
				}
				*/
				m_text = strdup(newtext);
				m_anchor = new_anchor;
				m_tempo = newtempo;
				m_enable_tempo_change = newenable_tempo;
				m_midi_volume = newvolume;
				m_enable_volume_change = newenable_volume;
				m_segno_sign = newsegno_sign;
				m_font_family = newfont_family;
				m_font_size = newfont_size;
				m_font_slant = newfont_slant;
				m_font_weight = newfont_weight;  
			}
		}
		computeBbox();
	
	}
}

void NedFreeText::computeBbox() {
	if (m_text == NULL) {
		m_bbox.x = m_bbox.y = m_bbox.width = m_bbox.height = 0.0;
		return;
	}
	cairo_text_extents_t extends;
	cairo_t *cr = gdk_cairo_create (m_drawing_area->window);

	cairo_select_font_face(cr, m_font_family, m_font_slant, m_font_weight);
	cairo_text_extents(cr, m_text , &extends);
	m_bbox.width = extends.width * m_font_size / FONT_REL_SIZE;
	m_bbox.height = extends.height * m_font_size / FONT_REL_SIZE;
	cairo_destroy(cr);
}

void NedFreeText::draw(cairo_t *cr, double leftx, double topy, double zoom_factor, int zoom_level) {
	if (m_element == NULL) return;

	double xpos, ypos;

	getXYPos(&xpos, &ypos);

	cairo_new_path(cr);
	if (isActive()) {
		cairo_set_source_rgb (cr, 1.0, 0.0, 0.0);
	}
	cairo_select_font_face(cr, m_font_family, m_font_slant, m_font_weight);
	cairo_set_font_size(cr, m_font_size * zoom_factor);
	switch (m_anchor) {
		case ANCHOR_LEFT: cairo_move_to(cr, X_POS_PAGE_REL(xpos), Y_POS_PAGE_REL(ypos)); break;
		case ANCHOR_MID: cairo_move_to(cr, X_POS_PAGE_REL(xpos - m_bbox.width / 2.0), Y_POS_PAGE_REL(ypos)); break;
		case ANCHOR_RIGHT: cairo_move_to(cr, X_POS_PAGE_REL(xpos - m_bbox.width), Y_POS_PAGE_REL(ypos)); break;
	}
	cairo_show_text(cr, m_text);
	cairo_stroke(cr);
//#define SHOW_BBOX
#ifdef SHOW_BBOX
	cairo_set_source_rgb (cr, 1.0, 0.0, 1.0);
	cairo_new_path(cr);
	cairo_set_line_width(cr, zoom_factor * STEM_THICK);
	cairo_rectangle (cr, X_POS_PAGE_REL(xpos), Y_POS_PAGE_REL(ypos - m_bbox.height), zoom_factor * (m_bbox.width), zoom_factor * m_bbox.height);
	cairo_stroke(cr);
	cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
#endif

	cairo_set_font_face(cr, NedResource::getFontFace());
	cairo_set_font_matrix(cr, NedResource::getFontMatrix(zoom_level));
	cairo_stroke(cr);
	if (isActive()) {
		cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
	}
	if (m_is_abs) {
		drawConnection(cr, xpos, ypos, leftx, topy, zoom_factor);
	}
}

bool NedFreeText::trySelect(double xl, double yl, double zoom_factor, double leftx, double topy) {
	if (m_element == NULL) return false;
	double xpos, ypos;
	double x = X_POS_PAGE_REL_INVERSE(xl);
	double y = Y_POS_PAGE_REL_INVERSE(yl);

	getXYPos(&xpos, &ypos);
	switch (m_anchor) {
		case ANCHOR_MID: xpos -= m_bbox.width / 2.0; break;
		case ANCHOR_RIGHT: xpos -= m_bbox.width; break;
	}
	/*
	if (x < xpos) {
		printf("raus wegen x(%f) < xpos(%f)\n", x, xpos); fflush(stdout);
		return false;
	}
	if (x > xpos + m_bbox.width) {
		printf("raus wegen x (%f) > xpos(%f) + m_bbox.width(%f) --> %f\n",
		x, xpos, m_bbox.width, xpos + m_bbox.width); fflush(stdout);
		return false;
	}
	if (y < ypos - m_bbox.height) {
		printf("raus wegen y(%f) < ypos(%f) - m_bbox.height(%f) -- %f\n",
			y, ypos, m_bbox.height, ypos - m_bbox.height); fflush(stdout);
		return false;
	}
	if (y > ypos) {
		printf("raus, wegen y(%f) > ypos(%f)\n", y, ypos); fflush(stdout);
		return false;
	}
	*/

	if (x < xpos || x > xpos + m_bbox.width || y < ypos - m_bbox.height || y > ypos) return false;
	setThis();
	return true;
}

double NedFreeText::getTempoInverse() {
	return 60000.0 / (double) m_tempo;
}

void NedFreeText::setText(char *text) {
	if (m_text != NULL) {
		free(m_text);
	}
	m_text = strdup(text);
}
